@@ -8,7 +8,7 @@
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 #
-# Copyright (c) 2004-2008 Kenneth J. Pronovici.
+# Copyright (c) 2004-2008,2010 Kenneth J. Pronovici.
 # All rights reserved.
 #
 # This program is free software; you can redistribute it and/or
@@ -25,9 +25,9 @@
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 #
 # Author   : Kenneth J. Pronovici <pronovic@ieee.org>
-# Language : Python (>= 2.3)
+# Language : Python (>= 2.5)
 # Project  : Cedar Backup, release 2
-# Revision : $Id: filesystem.py 879 2008-03-20 04:00:23Z pronovic $
+# Revision : $Id: filesystem.py 1080 2014-10-07 18:20:20Z pronovic $
 # Purpose  : Provides filesystem-related objects.
 #
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
@@ -48,18 +48,16 @@ Provides filesystem-related objects.
 ########################################################################
 
 # System modules
-import sys
 import os
 import re
-import sha
 import math
 import logging
 import tarfile
 
 # Cedar Backup modules
 from CedarBackup2.knapsack import firstFit, bestFit, worstFit, alternateFit
-from CedarBackup2.util import AbsolutePathList, ObjectTypeList, UnorderedList, RegexList
-from CedarBackup2.util import removeKeys, displayBytes, calculateFileAge, encodePath
+from CedarBackup2.util import AbsolutePathList, UnorderedList, RegexList
+from CedarBackup2.util import removeKeys, displayBytes, calculateFileAge, encodePath, dereferenceLink
 
 
 ########################################################################
@@ -115,7 +113,7 @@ class FilesystemList(list):
    because the operating system never reports a file as a soft link.
 
    @sort: __init__, addFile, addDir, addDirContents, removeFiles, removeDirs,
-          removeLinks, removeMatch, removeInvalid, normalize, validate, 
+          removeLinks, removeMatch, removeInvalid, normalize,
           excludeFiles, excludeDirs, excludeLinks, excludePaths, 
           excludePatterns, excludeBasenamePatterns, ignoreFile 
    """
@@ -203,15 +201,15 @@ class FilesystemList(list):
       Elements do not have to exist on disk at the time of assignment.
       @raise ValueError: If any list element is not an absolute path.
       """
-      self._absoluteExcludePaths = AbsolutePathList()
+      self._excludePaths = AbsolutePathList()
       if value is not None:
-         self._absoluteExcludePaths.extend(value)
+         self._excludePaths.extend(value)
 
    def _getExcludePaths(self):
       """
       Property target used to get the absolute exclude paths list.
       """
-      return self._absoluteExcludePaths
+      return self._excludePaths
 
    def _setExcludePatterns(self, value):
       """
@@ -304,10 +302,12 @@ class FilesystemList(list):
          logger.debug("Path [%s] is excluded based on excludePaths." % path)
          return 0
       for pattern in self.excludePatterns:
+         pattern = encodePath(pattern)  # use same encoding as filenames
          if re.compile(r"^%s$" % pattern).match(path): # safe to assume all are valid due to RegexList
             logger.debug("Path [%s] is excluded based on pattern [%s]." % (path, pattern))
             return 0
       for pattern in self.excludeBasenamePatterns: # safe to assume all are valid due to RegexList
+         pattern = encodePath(pattern)  # use same encoding as filenames
          if re.compile(r"^%s$" % pattern).match(os.path.basename(path)):
             logger.debug("Path [%s] is excluded based on basename pattern [%s]." % (path, pattern))
             return 0
@@ -347,10 +347,12 @@ class FilesystemList(list):
          logger.debug("Path [%s] is excluded based on excludePaths." % path)
          return 0
       for pattern in self.excludePatterns: # safe to assume all are valid due to RegexList
+         pattern = encodePath(pattern)  # use same encoding as filenames
          if re.compile(r"^%s$" % pattern).match(path):
             logger.debug("Path [%s] is excluded based on pattern [%s]." % (path, pattern))
             return 0
       for pattern in self.excludeBasenamePatterns: # safe to assume all are valid due to RegexList
+         pattern = encodePath(pattern)  # use same encoding as filenames
          if re.compile(r"^%s$" % pattern).match(os.path.basename(path)):
             logger.debug("Path [%s] is excluded based on basename pattern [%s]." % (path, pattern))
             return 0
@@ -358,7 +360,7 @@ class FilesystemList(list):
       logger.debug("Added directory to list: [%s]" % path)
       return 1
 
-   def addDirContents(self, path, recursive=True, addSelf=True, linkDepth=0):
+   def addDirContents(self, path, recursive=True, addSelf=True, linkDepth=0, dereference=False):
       """
       Adds the contents of a directory to the list.
 
@@ -387,6 +389,9 @@ class FilesystemList(list):
       @note: The L{excludeDirs} flag only controls whether any given directory
       path itself is added to the list once it has been discovered.  It does
       I{not} modify any behavior related to directory recursion.
+
+      @note: If you call this method I{on a link to a directory} that link will
+      never be dereferenced (it may, however, be followed).
    
       @param path: Directory path whose contents should be added to the list
       @type path: String representing a path on disk
@@ -400,6 +405,9 @@ class FilesystemList(list):
       @param linkDepth: Maximum depth of the tree at which soft links should be followed
       @type linkDepth: Integer value, where zero means not to follow any soft links
 
+      @param dereference: Indicates whether soft links, if followed, should be dereferenced
+      @type dereference: Boolean value
+
       @return: Number of items recursively added to the list
 
       @raise ValueError: If path is not a directory or does not exist.
@@ -407,9 +415,9 @@ class FilesystemList(list):
       """
       path = encodePath(path)
       path = normalizeDir(path)
-      return self._addDirContentsInternal(path, addSelf, recursive, linkDepth)
+      return self._addDirContentsInternal(path, addSelf, recursive, linkDepth, dereference)
 
-   def _addDirContentsInternal(self, path, includePath=True, recursive=True, linkDepth=0):
+   def _addDirContentsInternal(self, path, includePath=True, recursive=True, linkDepth=0, dereference=False):
       """
       Internal implementation of C{addDirContents}.
 
@@ -424,12 +432,23 @@ class FilesystemList(list):
       The linkDepth parameter controls whether soft links are followed when we
       are adding the contents recursively.  Any recursive calls reduce the
       value by one.  If the value zero or less, then soft links will just be
-      added as directories, but will not be followed.
+      added as directories, but will not be followed.  This means that links
+      are followed to a I{constant depth} starting from the top-most directory.
+
+      There is one difference between soft links and directories: soft links
+      that are added recursively are not placed into the list explicitly.  This
+      is because if we do add the links recursively, the resulting tar file
+      gets a little confused (it has a link and a directory with the same
+      name).
+
+      @note: If you call this method I{on a link to a directory} that link will
+      never be dereferenced (it may, however, be followed).
 
       @param path: Directory path whose contents should be added to the list.
       @param includePath: Indicates whether to include the path as well as contents.
       @param recursive: Indicates whether directory contents should be added recursively.
       @param linkDepth: Depth of soft links that should be followed
+      @param dereference: Indicates whether soft links, if followed, should be dereferenced
 
       @return: Number of items recursively added to the list
 
@@ -443,10 +462,12 @@ class FilesystemList(list):
          logger.debug("Path [%s] is excluded based on excludePaths." % path)
          return added
       for pattern in self.excludePatterns: # safe to assume all are valid due to RegexList
+         pattern = encodePath(pattern)  # use same encoding as filenames
          if re.compile(r"^%s$" % pattern).match(path):
             logger.debug("Path [%s] is excluded based on pattern [%s]." % (path, pattern))
             return added
       for pattern in self.excludeBasenamePatterns: # safe to assume all are valid due to RegexList
+         pattern = encodePath(pattern)  # use same encoding as filenames
          if re.compile(r"^%s$" % pattern).match(os.path.basename(path)):
             logger.debug("Path [%s] is excluded based on basename pattern [%s]." % (path, pattern))
             return added
@@ -458,18 +479,31 @@ class FilesystemList(list):
       for entry in os.listdir(path):
          entrypath = os.path.join(path, entry)
          if os.path.isfile(entrypath):
+            if linkDepth > 0 and dereference:
+               derefpath = dereferenceLink(entrypath)
+               if derefpath != entrypath:
+                  added += self.addFile(derefpath)
             added += self.addFile(entrypath)
          elif os.path.isdir(entrypath):
             if os.path.islink(entrypath):
-               if recursive and linkDepth > 0:
-                  newDepth = linkDepth - 1;
-                  added += self._addDirContentsInternal(entrypath, linkDepth=newDepth)
+               if recursive:
+                  if linkDepth > 0:
+                     newDepth = linkDepth - 1
+                     if dereference:
+                        derefpath = dereferenceLink(entrypath)
+                        if derefpath != entrypath:
+                           added += self._addDirContentsInternal(derefpath, True, recursive, newDepth, dereference)
+                        added += self.addDir(entrypath)
+                     else:
+                        added += self._addDirContentsInternal(entrypath, False, recursive, newDepth, dereference)
+                  else:
+                     added += self.addDir(entrypath)
                else:
                   added += self.addDir(entrypath)
             else:
                if recursive:
-                  newDepth = linkDepth - 1;
-                  added += self._addDirContentsInternal(entrypath, linkDepth=newDepth)
+                  newDepth = linkDepth - 1
+                  added += self._addDirContentsInternal(entrypath, True, recursive, newDepth, dereference)
                else:
                   added += self.addDir(entrypath)
       return added
@@ -507,6 +541,7 @@ class FilesystemList(list):
                removed += 1
       else:
          try:
+            pattern = encodePath(pattern)  # use same encoding as filenames
             compiled = re.compile(pattern)
          except re.error:
             raise ValueError("Pattern is not a valid regular expression.")
@@ -516,7 +551,7 @@ class FilesystemList(list):
                   self.remove(entry)
                   logger.debug("Removed path [%s] from list." % entry)
                   removed += 1
-      logger.debug("Removed a total of %d entries." % removed);
+      logger.debug("Removed a total of %d entries." % removed)
       return removed
 
    def removeDirs(self, pattern=None):
@@ -549,6 +584,7 @@ class FilesystemList(list):
                removed += 1
       else:
          try:
+            pattern = encodePath(pattern)  # use same encoding as filenames
             compiled = re.compile(pattern)
          except re.error:
             raise ValueError("Pattern is not a valid regular expression.")
@@ -558,7 +594,7 @@ class FilesystemList(list):
                   self.remove(entry)
                   logger.debug("Removed path [%s] from list based on pattern [%s]." % (entry, pattern))
                   removed += 1
-      logger.debug("Removed a total of %d entries." % removed);
+      logger.debug("Removed a total of %d entries." % removed)
       return removed
 
    def removeLinks(self, pattern=None):
@@ -589,6 +625,7 @@ class FilesystemList(list):
                removed += 1
       else:
          try:
+            pattern = encodePath(pattern)  # use same encoding as filenames
             compiled = re.compile(pattern)
          except re.error:
             raise ValueError("Pattern is not a valid regular expression.")
@@ -598,7 +635,7 @@ class FilesystemList(list):
                   self.remove(entry)
                   logger.debug("Removed path [%s] from list based on pattern [%s]." % (entry, pattern))
                   removed += 1
-      logger.debug("Removed a total of %d entries." % removed);
+      logger.debug("Removed a total of %d entries." % removed)
       return removed
 
    def removeMatch(self, pattern):
@@ -623,6 +660,7 @@ class FilesystemList(list):
       @raise ValueError: If the passed-in pattern is not a valid regular expression.
       """
       try:
+         pattern = encodePath(pattern)  # use same encoding as filenames
          compiled = re.compile(pattern)
       except re.error:
          raise ValueError("Pattern is not a valid regular expression.")
@@ -632,7 +670,7 @@ class FilesystemList(list):
             self.remove(entry)
             logger.debug("Removed path [%s] from list based on pattern [%s]." % (entry, pattern))
             removed += 1
-      logger.debug("Removed a total of %d entries." % removed);
+      logger.debug("Removed a total of %d entries." % removed)
       return removed
 
    def removeInvalid(self):
@@ -651,7 +689,7 @@ class FilesystemList(list):
             self.remove(entry)
             logger.debug("Removed path [%s] from list." % entry)
             removed += 1
-      logger.debug("Removed a total of %d entries." % removed);
+      logger.debug("Removed a total of %d entries." % removed)
       return removed
 
 
@@ -686,7 +724,7 @@ class FilesystemList(list):
 # SpanItem class definition
 ########################################################################
 
-class SpanItem(object):
+class SpanItem(object): # pylint: disable=R0903
    """
    Item returned by L{BackupFileList.generateSpan}.
    """
@@ -707,7 +745,7 @@ class SpanItem(object):
 # BackupFileList class definition
 ########################################################################
 
-class BackupFileList(FilesystemList):
+class BackupFileList(FilesystemList): # pylint: disable=R0904
 
    ######################
    # Class documentation
@@ -838,6 +876,7 @@ class BackupFileList(FilesystemList):
                table[entry] = BackupFileList._generateDigest(entry)
       return table
    
+   @staticmethod
    def _generateDigest(path):
       """
       Generates an SHA digest for a given file on disk.
@@ -859,7 +898,7 @@ class BackupFileList(FilesystemList):
       requires 111 seconds.  This implementation requires only 40-45 seconds,
       which is a pretty substantial speed-up.  
 
-      Practice shows that reading in around 4kB (4096 bytes) at a time yields
+      Experience shows that reading in around 4kB (4096 bytes) at a time yields
       the best performance.  Smaller reads are quite a bit slower, and larger
       reads don't make much of a difference.  The 4kB number makes me a little
       suspicious, and I think it might be related to the size of a filesystem
@@ -872,7 +911,13 @@ class BackupFileList(FilesystemList):
       @return: ASCII-safe SHA digest for the file.
       @raise OSError: If the file cannot be opened.
       """
-      s = sha.new()
+      # pylint: disable=C0103,E1101
+      try:
+         import hashlib
+         s = hashlib.sha1()
+      except ImportError:
+         import sha
+         s = sha.new()
       f = open(path, mode="rb")  # in case platform cares about binary reads
       readBytes = 4096  # see notes above
       while(readBytes > 0):
@@ -883,7 +928,6 @@ class BackupFileList(FilesystemList):
       digest = s.hexdigest()
       logger.debug("Generated digest [%s] for file [%s]." % (digest, path))
       return digest
-   _generateDigest = staticmethod(_generateDigest)
 
    def generateFitted(self, capacity, algorithm="worst_fit"):
       """
@@ -972,6 +1016,7 @@ class BackupFileList(FilesystemList):
             table[entry] = (entry, size)
       return table
 
+   @staticmethod
    def _getKnapsackFunction(algorithm):
       """
       Returns a reference to the function associated with an algorithm name.
@@ -989,8 +1034,7 @@ class BackupFileList(FilesystemList):
       elif algorithm == "alternate_fit":
          return alternateFit
       else:
-         raise ValueError("Algorithm [%s] is invalid." % algorithm);
-   _getKnapsackFunction = staticmethod(_getKnapsackFunction)
+         raise ValueError("Algorithm [%s] is invalid." % algorithm)
 
    def generateTarfile(self, path, mode='tar', ignore=False, flat=False):
       """
@@ -1053,6 +1097,7 @@ class BackupFileList(FilesystemList):
       @raise ValueError: If the path could not be encoded properly.
       @raise TarError: If there is a problem creating the tar file
       """
+      # pylint: disable=E1101
       path = encodePath(path)
       if len(self) == 0: raise ValueError("Empty list cannot be used to generate tarfile.")
       if(mode == 'tar'): tarmode = "w:"
@@ -1061,7 +1106,10 @@ class BackupFileList(FilesystemList):
       else: raise ValueError("Mode [%s] is not valid." % mode)
       try:
          tar = tarfile.open(path, tarmode)
-         tar.posix = False    # make a GNU-compatible archive without file length limits
+         try:
+            tar.format = tarfile.GNU_FORMAT
+         except AttributeError:
+            tar.posix = False
          for entry in self:
             try:
                if flat:
@@ -1180,7 +1228,7 @@ class BackupFileList(FilesystemList):
 # PurgeItemList class definition
 ########################################################################
 
-class PurgeItemList(FilesystemList):
+class PurgeItemList(FilesystemList): # pylint: disable=R0904
 
    ######################
    # Class documentation
@@ -1214,7 +1262,7 @@ class PurgeItemList(FilesystemList):
    # Add methods
    ##############
 
-   def addDirContents(self, path, recursive=True, addSelf=True, linkDepth=0):
+   def addDirContents(self, path, recursive=True, addSelf=True, linkDepth=0, dereference=False):
       """
       Adds the contents of a directory to the list.
 
@@ -1248,6 +1296,9 @@ class PurgeItemList(FilesystemList):
       path itself is added to the list once it has been discovered.  It does
       I{not} modify any behavior related to directory recursion.
 
+      @note: If you call this method I{on a link to a directory} that link will
+      never be dereferenced (it may, however, be followed).
+
       @param path: Directory path whose contents should be added to the list
       @type path: String representing a path on disk
 
@@ -1259,6 +1310,9 @@ class PurgeItemList(FilesystemList):
       @param linkDepth: Depth of soft links that should be followed
       @type linkDepth: Integer value, where zero means not to follow any soft links
 
+      @param dereference: Indicates whether soft links, if followed, should be dereferenced
+      @type dereference: Boolean value
+
       @return: Number of items recursively added to the list
 
       @raise ValueError: If path is not a directory or does not exist.
@@ -1266,7 +1320,7 @@ class PurgeItemList(FilesystemList):
       """
       path = encodePath(path)
       path = normalizeDir(path)
-      return super(PurgeItemList, self)._addDirContentsInternal(path, False, recursive, linkDepth)
+      return super(PurgeItemList, self)._addDirContentsInternal(path, False, recursive, linkDepth, dereference)
 
 
    ##################
